@@ -44,6 +44,7 @@
 #include "gtksourcemark.h"
 #include "gtksourcemarkssequence.h"
 #include "gtksourcesearchcontext.h"
+#include "gtksourcetag.h"
 #include "gtksourceview-i18n.h"
 #include "gtksourceview-typebuiltins.h"
 
@@ -140,10 +141,8 @@
 #define PROFILE(x)
 #endif
 
-/* For bracket matching */
-#define MAX_CHARS_BEFORE_FINDING_A_MATCH    10000
-
-#define CONTEXT_CLASSES_PREFIX "gtksourceview:context-classes:"
+#define BRACKET_MATCHING_CHARS_LIMIT	10000
+#define CONTEXT_CLASSES_PREFIX		"gtksourceview:context-classes:"
 
 /* Signals */
 enum
@@ -174,9 +173,7 @@ enum
 struct _GtkSourceBufferPrivate
 {
 	GtkTextTag *bracket_match_tag;
-	GtkTextMark *bracket_mark_cursor;
-	GtkTextMark *bracket_mark_match;
-	GtkSourceBracketMatchType bracket_match;
+	GtkSourceBracketMatchType bracket_match_state;
 
 	/* Hash table: category -> MarksSequence */
 	GHashTable *source_marks;
@@ -199,8 +196,6 @@ struct _GtkSourceBufferPrivate
 
 	guint highlight_syntax : 1;
 	guint highlight_brackets : 1;
-	guint constructed : 1;
-	guint allow_bracket_match : 1;
 	guint implicit_trailing_newline : 1;
 };
 
@@ -239,25 +234,22 @@ static void 	 gtk_source_buffer_real_mark_set	(GtkTextBuffer		 *buffer,
 							 const GtkTextIter	 *location,
 							 GtkTextMark		 *mark);
 
-static void 	 gtk_source_buffer_real_apply_tag	(GtkTextBuffer		 *buffer,
-							 GtkTextTag		 *tag,
-							 const GtkTextIter	 *start,
-							 const GtkTextIter	 *end);
-
 static void 	 gtk_source_buffer_real_mark_deleted	(GtkTextBuffer		 *buffer,
 							 GtkTextMark		 *mark);
 
 static void	 gtk_source_buffer_real_undo		(GtkSourceBuffer	 *buffer);
 static void	 gtk_source_buffer_real_redo		(GtkSourceBuffer	 *buffer);
 
+static void	 gtk_source_buffer_real_highlight_updated
+							(GtkSourceBuffer         *buffer,
+							 GtkTextIter             *start,
+							 GtkTextIter             *end);
+
 static void
 gtk_source_buffer_constructed (GObject *object)
 {
 	GtkSourceBuffer *buffer = GTK_SOURCE_BUFFER (object);
 
-	/* we need to know that the tag-table was set */
-	buffer->priv->constructed = TRUE;
-
 	if (buffer->priv->undo_manager == NULL)
 	{
 		/* This will install the default undo manager */
@@ -285,7 +277,6 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	text_buffer_class->insert_text = gtk_source_buffer_real_insert_text;
 	text_buffer_class->insert_pixbuf = gtk_source_buffer_real_insert_pixbuf;
 	text_buffer_class->insert_child_anchor = gtk_source_buffer_real_insert_anchor;
-	text_buffer_class->apply_tag = gtk_source_buffer_real_apply_tag;
 	text_buffer_class->mark_set = gtk_source_buffer_real_mark_set;
 	text_buffer_class->mark_deleted = gtk_source_buffer_real_mark_deleted;
 
@@ -413,7 +404,8 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	    g_signal_new_class_handler ("highlight_updated",
 	                                G_OBJECT_CLASS_TYPE (object_class),
 	                                G_SIGNAL_RUN_LAST,
-	                                NULL, NULL, NULL, NULL,
+	                                G_CALLBACK (gtk_source_buffer_real_highlight_updated),
+					NULL, NULL, NULL,
 	                                G_TYPE_NONE,
 	                                2,
 	                                GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE,
@@ -468,13 +460,18 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	/**
 	 * GtkSourceBuffer::bracket-matched:
 	 * @buffer: a #GtkSourceBuffer.
-	 * @iter: iterator to initialize.
-	 * @state: state of bracket matching
+	 * @iter: if found, the location of the matching bracket.
+	 * @state: state of bracket matching.
 	 *
-	 * Sets @iter to a valid iterator pointing to the matching bracket
-	 * if @state is #GTK_SOURCE_BRACKET_MATCH_FOUND. Otherwise @iter is
+	 * @iter is set to a valid iterator pointing to the matching bracket
+	 * if @state is %GTK_SOURCE_BRACKET_MATCH_FOUND. Otherwise @iter is
 	 * meaningless.
 	 *
+	 * The signal is emitted only when the @state changes, typically when
+	 * the cursor moves.
+	 *
+	 * A use-case for this signal is to show messages in a #GtkStatusbar.
+	 *
 	 * Since: 2.12
 	 */
 	buffer_signals[BRACKET_MATCHED] =
@@ -548,9 +545,7 @@ gtk_source_buffer_init (GtkSourceBuffer *buffer)
 
 	priv->highlight_syntax = TRUE;
 	priv->highlight_brackets = TRUE;
-	priv->bracket_mark_cursor = NULL;
-	priv->bracket_mark_match = NULL;
-	priv->bracket_match = GTK_SOURCE_BRACKET_MATCH_NONE;
+	priv->bracket_match_state = GTK_SOURCE_BRACKET_MATCH_NONE;
 	priv->max_undo_levels = -1;
 
 	priv->source_marks = g_hash_table_new_full (g_str_hash,
@@ -774,17 +769,19 @@ gtk_source_buffer_can_redo_handler (GtkSourceUndoManager *manager,
 static void
 update_bracket_match_style (GtkSourceBuffer *buffer)
 {
-	if (buffer->priv->bracket_match_tag != NULL)
+	GtkSourceStyle *style = NULL;
+
+	if (buffer->priv->bracket_match_tag == NULL)
 	{
-		GtkSourceStyle *style = NULL;
-
-		if (buffer->priv->style_scheme != NULL)
-		{
-			style = _gtk_source_style_scheme_get_matching_brackets_style (buffer->priv->style_scheme);
-		}
-
-		_gtk_source_style_apply (style, buffer->priv->bracket_match_tag);
+		return;
 	}
+
+	if (buffer->priv->style_scheme != NULL)
+	{
+		style = _gtk_source_style_scheme_get_matching_brackets_style (buffer->priv->style_scheme);
+	}
+
+	_gtk_source_style_apply (style, buffer->priv->bracket_match_tag);
 }
 
 static GtkTextTag *
@@ -819,7 +816,7 @@ bracket_pair (gunichar  base_char,
 	gint dir;
 	gunichar pair;
 
-	switch ((gint)base_char)
+	switch (base_char)
 	{
 		case '{':
 			dir = 1;
@@ -876,39 +873,23 @@ bracket_pair (gunichar  base_char,
 }
 
 static void
-gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
-			       const GtkTextIter *iter,
-			       GtkTextMark       *mark)
+update_bracket_highlighting (GtkSourceBuffer *source_buffer)
 {
-	GtkSourceBuffer *source_buffer;
-	GtkTextIter start, end;
-	gunichar cursor_char;
+	GtkTextBuffer *buffer;
+	GtkTextIter insert_iter;
+	GtkTextIter bracket;
+	GtkTextIter bracket_match;
 	GtkSourceBracketMatchType previous_state;
 
-	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
-	g_return_if_fail (iter != NULL);
-	g_return_if_fail (mark != NULL);
-	g_return_if_fail (gtk_text_iter_get_buffer (iter) == buffer);
+	buffer = GTK_TEXT_BUFFER (source_buffer);
 
-	if (mark != gtk_text_buffer_get_insert (buffer))
+	if (source_buffer->priv->bracket_match_tag != NULL)
 	{
-		return;
-	}
+		GtkTextIter start;
+		GtkTextIter end;
 
-	source_buffer = GTK_SOURCE_BUFFER (buffer);
+		gtk_text_buffer_get_bounds (buffer, &start, &end);
 
-	if (source_buffer->priv->bracket_match == GTK_SOURCE_BRACKET_MATCH_FOUND)
-	{
-		gtk_text_buffer_get_iter_at_mark (buffer,
-						  &start,
-						  source_buffer->priv->bracket_mark_match);
-
-		gtk_text_buffer_get_iter_at_mark (buffer,
-						  &end,
-						  source_buffer->priv->bracket_mark_cursor);
-
-		gtk_text_iter_order (&start, &end);
-		gtk_text_iter_forward_char (&end);
 		gtk_text_buffer_remove_tag (buffer,
 					    source_buffer->priv->bracket_match_tag,
 					    &start,
@@ -917,109 +898,94 @@ gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 
 	if (!source_buffer->priv->highlight_brackets)
 	{
+		if (source_buffer->priv->bracket_match_tag != NULL)
+		{
+			GtkTextTagTable *table;
+
+			table = gtk_text_buffer_get_tag_table (buffer);
+			gtk_text_tag_table_remove (table, source_buffer->priv->bracket_match_tag);
+			source_buffer->priv->bracket_match_tag = NULL;
+		}
+
 		return;
 	}
 
-	start = *iter;
-	previous_state = source_buffer->priv->bracket_match;
-	source_buffer->priv->bracket_match = _gtk_source_buffer_find_bracket_match (source_buffer, &start);
+	gtk_text_buffer_get_iter_at_mark (buffer,
+					  &insert_iter,
+					  gtk_text_buffer_get_insert (buffer));
 
-	if (source_buffer->priv->bracket_match != GTK_SOURCE_BRACKET_MATCH_FOUND)
+	previous_state = source_buffer->priv->bracket_match_state;
+	source_buffer->priv->bracket_match_state =
+		_gtk_source_buffer_find_bracket_match (source_buffer,
+						       &insert_iter,
+						       &bracket,
+						       &bracket_match);
+
+	if (source_buffer->priv->bracket_match_state == GTK_SOURCE_BRACKET_MATCH_FOUND)
 	{
-		/* don't emit the signal at all if chars at previous and current
-		   positions are nonbrackets. */
-		if (previous_state != GTK_SOURCE_BRACKET_MATCH_NONE ||
-		    source_buffer->priv->bracket_match != GTK_SOURCE_BRACKET_MATCH_NONE)
-		{
-			g_signal_emit (source_buffer,
-				       buffer_signals[BRACKET_MATCHED],
-				       0,
-				       &end,
-				       source_buffer->priv->bracket_match);
-		}
+		GtkTextIter next_iter;
+
+		g_signal_emit (source_buffer,
+			       buffer_signals[BRACKET_MATCHED],
+			       0,
+			       &bracket_match,
+			       GTK_SOURCE_BRACKET_MATCH_FOUND);
+
+		next_iter = bracket_match;
+		gtk_text_iter_forward_char (&next_iter);
+		gtk_text_buffer_apply_tag (buffer,
+					   get_bracket_match_tag (source_buffer),
+					   &bracket_match,
+					   &next_iter);
+
+		next_iter = bracket;
+		gtk_text_iter_forward_char (&next_iter);
+		gtk_text_buffer_apply_tag (buffer,
+					   get_bracket_match_tag (source_buffer),
+					   &bracket,
+					   &next_iter);
+		return;
 	}
-	else
+
+	/* Don't emit the signal at all if chars at previous and current
+	 * positions are nonbrackets.
+	 */
+	if (previous_state != GTK_SOURCE_BRACKET_MATCH_NONE ||
+	    source_buffer->priv->bracket_match_state != GTK_SOURCE_BRACKET_MATCH_NONE)
 	{
 		g_signal_emit (source_buffer,
 			       buffer_signals[BRACKET_MATCHED],
 			       0,
-			       &start,
-			       source_buffer->priv->bracket_match);
-
-		/* allow_bracket_match will allow the bracket match tag to be
-		   applied to the buffer. See apply_tag_real for more
-		   information */
-		source_buffer->priv->allow_bracket_match = TRUE;
-
-		/* Mark matching bracket */
-		if (!source_buffer->priv->bracket_mark_match)
-		{
-			source_buffer->priv->bracket_mark_match =
- 				gtk_text_buffer_create_mark (buffer,
-							     NULL,
-							     &start,
-							     TRUE);
- 		}
- 		else
- 		{
- 			gtk_text_buffer_move_mark (buffer,
-						   source_buffer->priv->bracket_mark_match,
-						   &start);
- 		}
-
-		end = start;
-		gtk_text_iter_forward_char (&end);
-		gtk_text_buffer_apply_tag (buffer,
-					   get_bracket_match_tag (source_buffer),
-					   &start,
-					   &end);
-
-		/* Mark the bracket near the cursor */
-		start = *iter;
-		cursor_char = gtk_text_iter_get_char (&start);
-		if (bracket_pair (cursor_char, NULL) == 0)
-		{
-			gtk_text_iter_backward_char (&start);
-		}
-
-		if (!source_buffer->priv->bracket_mark_cursor)
-		{
-			source_buffer->priv->bracket_mark_cursor =
-				gtk_text_buffer_create_mark (buffer,
-							     NULL,
-							     &start,
-							     FALSE);
-		}
-		else
-		{
-			gtk_text_buffer_move_mark (buffer,
-						   source_buffer->priv->bracket_mark_cursor,
-						   &start);
-		}
-
-		end = start;
-		gtk_text_iter_forward_char (&end);
-		gtk_text_buffer_apply_tag (buffer,
-					   get_bracket_match_tag (source_buffer),
-					   &start,
-					   &end);
-
-		source_buffer->priv->allow_bracket_match = FALSE;
+			       NULL,
+			       source_buffer->priv->bracket_match_state);
 	}
 }
 
+/* Although this function is not really useful (update_bracket_highlighting()
+ * could be called directly), the name cursor_moved() is more meaningful.
+ */
+static void
+cursor_moved (GtkSourceBuffer *buffer)
+{
+	update_bracket_highlighting (buffer);
+}
+
+static void
+gtk_source_buffer_real_highlight_updated (GtkSourceBuffer *buffer,
+					  GtkTextIter     *start,
+					  GtkTextIter     *end)
+{
+	update_bracket_highlighting (buffer);
+}
+
 static void
 gtk_source_buffer_content_inserted (GtkTextBuffer *buffer,
 				    gint           start_offset,
 				    gint           end_offset)
 {
-	GtkTextMark *mark;
-	GtkTextIter insert_iter;
 	GtkSourceBuffer *source_buffer = GTK_SOURCE_BUFFER (buffer);
 
-	mark = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &insert_iter, mark);
-	gtk_source_buffer_move_cursor (buffer, &insert_iter, mark);
+	cursor_moved (source_buffer);
 
 	if (source_buffer->priv->highlight_engine != NULL)
 	{
@@ -1115,8 +1081,6 @@ gtk_source_buffer_real_delete_range (GtkTextBuffer *buffer,
 				     GtkTextIter   *end)
 {
 	gint offset, length;
-	GtkTextMark *mark;
-	GtkTextIter iter;
 	GtkSourceBuffer *source_buffer = GTK_SOURCE_BUFFER (buffer);
 
 	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
@@ -1131,9 +1095,7 @@ gtk_source_buffer_real_delete_range (GtkTextBuffer *buffer,
 
 	GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->delete_range (buffer, start, end);
 
-	mark = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &iter, mark);
-	gtk_source_buffer_move_cursor (buffer, &iter, mark);
+	cursor_moved (source_buffer);
 
 	/* emit text deleted for engines */
 	if (source_buffer->priv->highlight_engine != NULL)
@@ -1172,89 +1134,88 @@ get_bracket_matching_context_class_mask (GtkSourceBuffer *buffer,
 	return mask;
 }
 
-/* Note that we only look MAX_CHARS_BEFORE_FINDING_A_MATCH at most. */
+/* Note that we only look BRACKET_MATCHING_CHARS_LIMIT at most.
+ * @pos is moved to the bracket match, if found.
+ */
 static GtkSourceBracketMatchType
-gtk_source_buffer_find_bracket_match_real (GtkSourceBuffer *buffer,
-                                           GtkTextIter     *orig)
+find_bracket_match_real (GtkSourceBuffer *buffer,
+			 GtkTextIter     *pos)
 {
 	GtkTextIter iter;
-
 	gunichar base_char;
 	gunichar search_char;
-	gunichar cur_char;
-	gint addition;
-	gint char_cont;
-	gint counter;
-
+	gint direction;
+	gint bracket_count;
+	gint char_count;
+	gint cclass_mask;
 	gboolean found;
 
-	gint cclass_mask;
+	base_char = gtk_text_iter_get_char (pos);
+	search_char = bracket_pair (base_char, &direction);
 
-	iter = *orig;
-
-	cur_char = gtk_text_iter_get_char (&iter);
-
-	base_char = cur_char;
-	cclass_mask = get_bracket_matching_context_class_mask (buffer, &iter);
-
-	search_char = bracket_pair (base_char, &addition);
-
-	if (addition == 0)
+	if (direction == 0)
 	{
 		return GTK_SOURCE_BRACKET_MATCH_NONE;
 	}
 
-	counter = 0;
+	cclass_mask = get_bracket_matching_context_class_mask (buffer, pos);
+
+	iter = *pos;
+	bracket_count = 0;
+	char_count = 0;
 	found = FALSE;
-	char_cont = 0;
 
 	do
 	{
-		gint current_mask;
+		gunichar cur_char;
+		gint cur_mask;
 
-		gtk_text_iter_forward_chars (&iter, addition);
+		gtk_text_iter_forward_chars (&iter, direction);
 		cur_char = gtk_text_iter_get_char (&iter);
-		++char_cont;
+		char_count++;
 
-		current_mask = get_bracket_matching_context_class_mask (buffer, &iter);
+		cur_mask = get_bracket_matching_context_class_mask (buffer, &iter);
 
 		/* Check if we lost a class, which means we don't look any
-		   further */
-		if (current_mask < cclass_mask)
+		 * further.
+		 */
+		if ((cclass_mask & cur_mask) != cclass_mask)
 		{
 			found = FALSE;
 			break;
 		}
 
-		if ((cur_char == search_char || cur_char == base_char) &&
-		    cclass_mask == current_mask)
+		if (cclass_mask != cur_mask)
 		{
-			if ((cur_char == search_char) && counter == 0)
+			continue;
+		}
+
+		if (cur_char == search_char)
+		{
+			if (bracket_count == 0)
 			{
 				found = TRUE;
 				break;
 			}
 
-			if (cur_char == base_char)
-			{
-				counter++;
-			}
-			else
-			{
-				counter--;
-			}
+			bracket_count--;
+		}
+		else if (cur_char == base_char)
+		{
+			bracket_count++;
 		}
 	}
-	while (!gtk_text_iter_is_end (&iter) && !gtk_text_iter_is_start (&iter) &&
-		(char_cont < MAX_CHARS_BEFORE_FINDING_A_MATCH));
+	while (!gtk_text_iter_is_end (&iter) &&
+	       !gtk_text_iter_is_start (&iter) &&
+	       char_count < BRACKET_MATCHING_CHARS_LIMIT);
 
 	if (found)
 	{
-		*orig = iter;
+		*pos = iter;
 		return GTK_SOURCE_BRACKET_MATCH_FOUND;
 	}
 
-	if (char_cont >= MAX_CHARS_BEFORE_FINDING_A_MATCH)
+	if (char_count >= BRACKET_MATCHING_CHARS_LIMIT)
 	{
 		return GTK_SOURCE_BRACKET_MATCH_OUT_OF_RANGE;
 	}
@@ -1262,36 +1223,77 @@ gtk_source_buffer_find_bracket_match_real (GtkSourceBuffer *buffer,
 	return GTK_SOURCE_BRACKET_MATCH_NOT_FOUND;
 }
 
-/* Note that we take into account both the character following the cursor and the
- * one preceding it. If there are brackets on both sides the one following the
- * cursor takes precedence.
+/* Note that we take into account both the character following @pos and the one
+ * preceding it. If there are brackets on both sides, the one following @pos
+ * takes precedence.
+ * @bracket and @bracket_match are valid only if GTK_SOURCE_BRACKET_MATCH_FOUND
+ * is returned. @bracket is set either to @pos or @pos-1.
  */
 GtkSourceBracketMatchType
-_gtk_source_buffer_find_bracket_match (GtkSourceBuffer *buffer,
-                                       GtkTextIter     *orig)
+_gtk_source_buffer_find_bracket_match (GtkSourceBuffer   *buffer,
+				       const GtkTextIter *pos,
+				       GtkTextIter       *bracket,
+				       GtkTextIter       *bracket_match)
 {
-	GtkSourceBracketMatchType res;
+	GtkSourceBracketMatchType result_right;
+	GtkSourceBracketMatchType result_left;
+	GtkTextIter prev;
 
-	res = gtk_source_buffer_find_bracket_match_real (buffer, orig);
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), GTK_SOURCE_BRACKET_MATCH_NONE);
+	g_return_val_if_fail (pos != NULL, GTK_SOURCE_BRACKET_MATCH_NONE);
+	g_return_val_if_fail (bracket_match != NULL, GTK_SOURCE_BRACKET_MATCH_NONE);
 
-	if (res != GTK_SOURCE_BRACKET_MATCH_FOUND)
+	*bracket_match = *pos;
+	result_right = find_bracket_match_real (buffer, bracket_match);
+
+	if (result_right == GTK_SOURCE_BRACKET_MATCH_FOUND)
 	{
-		GtkTextIter iter;
-		iter = *orig;
-
-		if (!gtk_text_iter_starts_line (&iter) &&
-		    gtk_text_iter_backward_char (&iter))
+		if (bracket != NULL)
 		{
-			res = gtk_source_buffer_find_bracket_match_real (buffer, &iter);
-
-			if (res == GTK_SOURCE_BRACKET_MATCH_FOUND)
-			{
-				*orig = iter;
-			}
+			*bracket = *pos;
 		}
+
+		return GTK_SOURCE_BRACKET_MATCH_FOUND;
 	}
 
-	return res;
+	prev = *pos;
+	if (!gtk_text_iter_starts_line (&prev) &&
+	    gtk_text_iter_backward_cursor_position (&prev))
+	{
+		*bracket_match = prev;
+		result_left = find_bracket_match_real (buffer, bracket_match);
+	}
+	else
+	{
+		result_left = GTK_SOURCE_BRACKET_MATCH_NONE;
+	}
+
+	if (result_left == GTK_SOURCE_BRACKET_MATCH_FOUND)
+	{
+		if (bracket != NULL)
+		{
+			*bracket = prev;
+		}
+
+		return GTK_SOURCE_BRACKET_MATCH_FOUND;
+	}
+
+	/* If there is a bracket, the expected return value is for the bracket,
+	 * not the other character.
+	 */
+	if (result_right == GTK_SOURCE_BRACKET_MATCH_NONE)
+	{
+		return result_left;
+	}
+	if (result_left == GTK_SOURCE_BRACKET_MATCH_NONE)
+	{
+		return result_right;
+	}
+
+	/* There are brackets on both sides, and none was successful. The one on
+	 * the right takes precedence.
+	 */
+	return result_right;
 }
 
 /**
@@ -1387,7 +1389,9 @@ gtk_source_buffer_get_max_undo_levels (GtkSourceBuffer *buffer)
  * track.  If the number of user actions exceeds the limit set by this
  * function, older actions will be discarded.
  *
- * If @max_undo_levels is -1, no limit is set.
+ * If @max_undo_levels is -1, the undo/redo is unlimited.
+ *
+ * If @max_undo_levels is 0, the undo/redo is disabled.
  */
 void
 gtk_source_buffer_set_max_undo_levels (GtkSourceBuffer *buffer,
@@ -1411,6 +1415,25 @@ gtk_source_buffer_set_max_undo_levels (GtkSourceBuffer *buffer,
 	g_object_notify (G_OBJECT (buffer), "max-undo-levels");
 }
 
+gboolean
+_gtk_source_buffer_is_undo_redo_enabled (GtkSourceBuffer *buffer)
+{
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), FALSE);
+
+	if (buffer->priv->undo_manager == NULL)
+	{
+		return FALSE;
+	}
+
+	/* A custom UndoManager is not forced to follow max_undo_levels. */
+	if (!GTK_SOURCE_IS_UNDO_MANAGER_DEFAULT (buffer->priv->undo_manager))
+	{
+		return TRUE;
+	}
+
+	return buffer->priv->max_undo_levels != 0;
+}
+
 /**
  * gtk_source_buffer_begin_not_undoable_action:
  * @buffer: a #GtkSourceBuffer.
@@ -1486,21 +1509,9 @@ gtk_source_buffer_set_highlight_matching_brackets (GtkSourceBuffer *buffer,
 
 	if (highlight != buffer->priv->highlight_brackets)
 	{
-		GtkTextIter iter;
-		GtkTextMark *mark;
-
 		buffer->priv->highlight_brackets = highlight;
 
-		/* try to see if there is already a bracket match at the
-		 * current position, but only if the tag table is already set
-		 * otherwise we have problems when calling this function
-		 * on init (get_insert creates the tag table as a side effect */
-		if (buffer->priv->constructed)
-		{
-			mark = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (buffer));
-			gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (buffer), &iter, mark);
-			gtk_source_buffer_move_cursor (GTK_TEXT_BUFFER (buffer), &iter, mark);
-		}
+		update_bracket_highlighting (buffer);
 
 		g_object_notify (G_OBJECT (buffer), "highlight-matching-brackets");
 	}
@@ -1752,30 +1763,6 @@ gtk_source_buffer_get_style_scheme (GtkSourceBuffer *buffer)
 	return buffer->priv->style_scheme;
 }
 
-static void
-gtk_source_buffer_real_apply_tag (GtkTextBuffer     *text_buffer,
-                                  GtkTextTag        *tag,
-                                  const GtkTextIter *start,
-                                  const GtkTextIter *end)
-{
-	GtkSourceBuffer *source_buffer = GTK_SOURCE_BUFFER (text_buffer);
-
-	/* We only allow the bracket match tag to be applied when we are doing
-	 * it ourselves (i.e. when allow_bracket_match is TRUE). The reason for
-	 * doing so is that when you copy/paste from the same buffer, the tags
-	 * get pasted too. This is ok for highlighting because the region will
-	 * get rehighlighted, but not for bracket matching.
-	 */
-	if (source_buffer->priv->allow_bracket_match ||
-	    tag != get_bracket_match_tag (source_buffer))
-	{
-		GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->apply_tag (text_buffer,
-										   tag,
-										   start,
-										   end);
-	}
-}
-
 static void
 add_source_mark (GtkSourceBuffer *buffer,
 		 GtkSourceMark   *mark)
@@ -1813,11 +1800,9 @@ gtk_source_buffer_real_mark_set	(GtkTextBuffer     *buffer,
 
 		g_signal_emit (buffer, buffer_signals[SOURCE_MARK_UPDATED], 0, mark);
 	}
-
-	/* If the mark is the insert mark, update bracket matching. */
 	else if (mark == gtk_text_buffer_get_insert (buffer))
 	{
-		gtk_source_buffer_move_cursor (buffer, location, mark);
+		cursor_moved (GTK_SOURCE_BUFFER (buffer));
 	}
 
 	GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->mark_set (buffer, location, mark);
@@ -3096,3 +3081,62 @@ gtk_source_buffer_get_implicit_trailing_newline (GtkSourceBuffer *buffer)
 
 	return buffer->priv->implicit_trailing_newline;
 }
+
+/**
+ * gtk_source_buffer_create_source_tag:
+ * @buffer: a #GtkSourceBuffer
+ * @tag_name: (nullable): name of the new tag, or %NULL
+ * @first_property_name: (nullable): name of first property to set, or %NULL
+ * @...: %NULL-terminated list of property names and values
+ *
+ * In short, this is the same function as gtk_text_buffer_create_tag(), but
+ * instead of creating a #GtkTextTag, this function creates a #GtkSourceTag.
+ *
+ * This function creates a #GtkSourceTag and adds it to the tag table for
+ * @buffer.  Equivalent to calling gtk_text_tag_new() and then adding the tag to
+ * the buffer’s tag table. The returned tag is owned by the buffer’s tag table,
+ * so the ref count will be equal to one.
+ *
+ * If @tag_name is %NULL, the tag is anonymous.
+ *
+ * If @tag_name is non-%NULL, a tag called @tag_name must not already
+ * exist in the tag table for this buffer.
+ *
+ * The @first_property_name argument and subsequent arguments are a list
+ * of properties to set on the tag, as with g_object_set().
+ *
+ * Returns: (transfer none): a new #GtkSourceTag.
+ * Since: 3.20
+ */
+GtkTextTag *
+gtk_source_buffer_create_source_tag (GtkSourceBuffer *buffer,
+				     const gchar     *tag_name,
+				     const gchar     *first_property_name,
+				     ...)
+{
+	GtkTextTag *tag;
+	GtkTextTagTable *table;
+	va_list list;
+
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), NULL);
+
+	tag = gtk_source_tag_new (tag_name);
+
+	table = gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (buffer));
+	if (!gtk_text_tag_table_add (table, tag))
+	{
+		g_object_unref (tag);
+		return NULL;
+	}
+
+	if (first_property_name != NULL)
+	{
+		va_start (list, first_property_name);
+		g_object_set_valist (G_OBJECT (tag), first_property_name, list);
+		va_end (list);
+	}
+
+	g_object_unref (tag);
+
+	return tag;
+}
