@@ -4,19 +4,18 @@
  * Copyright (C) 2015 Christian Hergert <christian@hergert.me>
  * Copyright (C) 2015 Ignacio Casal Quinteiro <icq@gnome.org>
  *
- * GtkSourceView is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
- * GtkSourceView is distributed in the hope that it will be useful,
+ * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "gtksourcemap.h"
@@ -301,7 +300,7 @@ gtk_source_map_rebuild_css (GtkSourceMap *map)
 		gchar *css;
 
 		css = _gtk_source_pango_font_description_to_css (priv->font_desc);
-		g_string_append_printf (gstr, "GtkSourceMap { %s }\n", css != NULL ? css : "");
+		g_string_append_printf (gstr, "textview { %s }\n", css != NULL ? css : "");
 		g_free (css);
 	}
 
@@ -347,11 +346,8 @@ gtk_source_map_rebuild_css (GtkSourceMap *map)
 		context = gtk_widget_get_style_context (GTK_WIDGET (priv->view));
 		gtk_style_context_save (context);
 		gtk_style_context_add_class (context, "view");
-		gtk_style_context_set_state (context, GTK_STATE_FLAG_SELECTED);
 		G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
-		gtk_style_context_get_background_color (context,
-							gtk_style_context_get_state (context),
-							&color);
+		gtk_style_context_get_background_color (context, GTK_STATE_FLAG_SELECTED, &color);
 		G_GNUC_END_IGNORE_DEPRECATIONS;
 		gtk_style_context_restore (context);
 		background = gdk_rgba_to_string (&color);
@@ -378,7 +374,7 @@ gtk_source_map_rebuild_css (GtkSourceMap *map)
 	if (background != NULL)
 	{
 		g_string_append_printf (gstr,
-		                        "GtkSourceMap.scrubber {\n"
+		                        "textview.scrubber {\n"
 		                        "\tbackground-color: %s;\n"
 		                        "\tborder-top: 1px solid shade(%s,0.9);\n"
 		                        "\tborder-bottom: 1px solid shade(%s,0.9);\n"
@@ -732,13 +728,6 @@ connect_view (GtkSourceMap  *map,
 		gtk_widget_add_events (GTK_WIDGET (priv->view), GDK_LEAVE_NOTIFY_MASK);
 	}
 
-	/* If we are not visible, we want to block certain signal handlers */
-	if (!gtk_widget_get_visible (GTK_WIDGET (map)))
-	{
-		g_signal_handler_block (vadj, priv->view_vadj_value_changed_handler);
-		g_signal_handler_block (vadj, priv->view_vadj_notify_upper_handler);
-	}
-
 	gtk_source_map_rebuild_css (map);
 }
 
@@ -1061,48 +1050,6 @@ gtk_source_map_realize (GtkWidget *widget)
 	set_view_cursor (GTK_SOURCE_MAP (widget));
 }
 
-static void
-gtk_source_map_show (GtkWidget *widget)
-{
-	GtkSourceMap *map = GTK_SOURCE_MAP (widget);
-	GtkSourceMapPrivate *priv;
-	GtkAdjustment *vadj;
-
-	GTK_WIDGET_CLASS (gtk_source_map_parent_class)->show (widget);
-
-	priv = gtk_source_map_get_instance_private (map);
-
-	if (priv->view != NULL)
-	{
-		vadj = gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (priv->view));
-
-		g_signal_handler_unblock (vadj, priv->view_vadj_value_changed_handler);
-		g_signal_handler_unblock (vadj, priv->view_vadj_notify_upper_handler);
-
-		g_object_notify (G_OBJECT (vadj), "upper");
-		g_signal_emit_by_name (vadj, "value-changed");
-	}
-}
-
-static void
-gtk_source_map_hide (GtkWidget *widget)
-{
-	GtkSourceMap *map = GTK_SOURCE_MAP (widget);
-	GtkSourceMapPrivate *priv;
-	GtkAdjustment *vadj;
-
-	GTK_WIDGET_CLASS (gtk_source_map_parent_class)->hide (widget);
-
-	priv = gtk_source_map_get_instance_private (map);
-
-	if (priv->view != NULL)
-	{
-		vadj = gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (priv->view));
-		g_signal_handler_block (vadj, priv->view_vadj_value_changed_handler);
-		g_signal_handler_block (vadj, priv->view_vadj_notify_upper_handler);
-	}
-}
-
 static void
 gtk_source_map_class_init (GtkSourceMapClass *klass)
 {
@@ -1116,13 +1063,11 @@ gtk_source_map_class_init (GtkSourceMapClass *klass)
 	widget_class->draw = gtk_source_map_draw;
 	widget_class->get_preferred_height = gtk_source_map_get_preferred_height;
 	widget_class->get_preferred_width = gtk_source_map_get_preferred_width;
-	widget_class->hide = gtk_source_map_hide;
 	widget_class->size_allocate = gtk_source_map_size_allocate;
 	widget_class->button_press_event = gtk_source_map_button_press_event;
 	widget_class->button_release_event = gtk_source_map_button_release_event;
 	widget_class->motion_notify_event = gtk_source_map_motion_notify_event;
 	widget_class->scroll_event = gtk_source_map_scroll_event;
-	widget_class->show = gtk_source_map_show;
 	widget_class->state_flags_changed = gtk_source_map_state_flags_changed;
 	widget_class->realize = gtk_source_map_realize;
 
