@@ -674,16 +674,11 @@ get_cursors_css_style (GtkSourceStyleScheme *scheme,
 
 		context = gtk_widget_get_style_context (widget);
 
-		gtk_style_context_save (context);
-		gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
-
 		gtk_style_context_get (context,
-				       gtk_style_context_get_state (context),
+				       GTK_STATE_FLAG_NORMAL,
 				       "background-color", &rgba,
 				       NULL);
 
-		gtk_style_context_restore (context);
-
 		/* shade the secondary cursor */
 		secondary_color.red = rgba->red * 0.5;
 		secondary_color.green = rgba->green * 0.5;
@@ -695,7 +690,7 @@ get_cursors_css_style (GtkSourceStyleScheme *scheme,
 	primary_color_str = gdk_rgba_to_string (&primary_color);
 	secondary_color_str = gdk_rgba_to_string (&secondary_color);
 
-	css = g_strdup_printf ("GtkSourceView {\n"
+	css = g_strdup_printf ("textview {\n"
 			       "\t-GtkWidget-cursor-color: %s;\n"
 			       "\t-GtkWidget-secondary-cursor-color: %s;\n"
 			       "}\n",
@@ -899,29 +894,29 @@ generate_css_style (GtkSourceStyleScheme *scheme)
 	final_style = g_string_new ("");
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_TEXT);
-	append_css_style (final_style, style, ".view");
+	append_css_style (final_style, style, "text, .view");
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED);
-	append_css_style (final_style, style, ".view:selected:focused");
+	append_css_style (final_style, style, "textview:selected:focused");
 
 	style2 = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED_UNFOCUSED);
 	append_css_style (final_style,
 			  style2 != NULL ? style2 : style,
-			  ".view:selected");
+			  "textview:selected");
 
 	/* For now we use "line numbers" colors for all the gutters */
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_LINE_NUMBERS);
 	if (style != NULL)
 	{
-		append_css_style (final_style, style, ".top");
-		append_css_style (final_style, style, ".right");
-		append_css_style (final_style, style, ".bottom");
-		append_css_style (final_style, style, ".left");
+		append_css_style (final_style, style, "border.top");
+		append_css_style (final_style, style, "border.right");
+		append_css_style (final_style, style, "border.bottom");
+		append_css_style (final_style, style, "border.left");
 
 		/* For the corners if the top or bottom gutter is also
 		 * displayed.
 		 */
-		append_css_style (final_style, style, "GtkSourceView");
+		append_css_style (final_style, style, "textview");
 	}
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_CURRENT_LINE_NUMBER);
